@@ -172,7 +172,7 @@ def metastable_from_msm(msm, n_hidden_states: int,
         equilibrium distribution.
     separate_symbols : array_like, optional, default=None
         Force the given set of observed states to stay in a separate hidden state.
-        The remaining nstates-1 states will be assigned by a metastable decomposition.
+        The remaining `#(observed states)-1` states will be assigned by a metastable decomposition.
     regularize : bool, optional, default=True
         If set to True, makes sure that the hidden initial distribution and transition matrix have nonzero probabilities
         by setting them to eps and then renormalizing. Avoids zeros that would cause estimation algorithms to crash or
@@ -232,6 +232,8 @@ def metastable_from_msm(msm, n_hidden_states: int,
                          np.eye(n_meta))
     else:
         pcca = nonseparate_msm.pcca(n_meta)
+
+    separate_states = None
     if separate_symbols is not None:
         separate_states = msm.count_model.symbols_to_states(separate_symbols)
         memberships = np.zeros((msm.n_states, n_hidden_states))
@@ -239,7 +241,6 @@ def metastable_from_msm(msm, n_hidden_states: int,
         memberships[separate_states, -1] = 1
     else:
         memberships = pcca.memberships
-        separate_states = None
 
     hidden_transition_matrix = _coarse_grain_transition_matrix(msm.transition_matrix, memberships)
     if reversible:
@@ -333,7 +334,7 @@ def metastable_from_data(dtrajs, n_hidden_states, lagtime, stride=1, mode='large
         equilibrium distribution.
     separate_symbols : array_like, optional, default=None
         Force the given set of observed states to stay in a separate hidden state.
-        The remaining nstates-1 states will be assigned by a metastable decomposition.
+        The remaining `#(observed states)-1` states will be assigned by a metastable decomposition.
     states : (dtype=int) ndarray, optional, default=None
         Artifically restrict count model to selection of states, even before regularization.
     regularize : bool, optional, default=True
